// dynamic light marker at textures in rl_gldefs
class LightActor : Actor
{
	Default
	{
		Scale 0.0;
		+FLOAT
		+NOGRAVITY
		+NOINTERACTION
		+NOBLOCKMAP
	}

	override void beginPlay()
	{
		super.beginPlay();
		changeStatNum(STAT_LightActor);
	}

	override void postBeginPlay()
	{
		if (rlClass_CVars.get().rl_debug.getBool())
		{
			self.Scale = (0.5, 0.5);
		}
	}

	override void tick() { }

	States
	{
	Spawn:
		APBX E -1 Bright;
		Stop;
	}
}
class BrightLightActor : LightActor
{
}

// bright light marker at 2-sided lines
class BrightLight : Actor
{
	Default
	{
		Scale 0.0;
		+FLOAT
		+NOGRAVITY
		+NOINTERACTION
		+NOBLOCKMAP
	}

	override void beginPlay()
	{
		super.beginPlay();
		changeStatNum(STAT_BrightLight);
	}

	override void postBeginPlay()
	{
		if (rlClass_CVars.get().rl_debug.getBool())
		{
			self.Scale = (0.5, 0.5);
		}
	}

	override void tick() { }

	States
	{
	Spawn:
		BAL1 A -1 Bright;
		Stop;
	}
}

class LightSensor : Actor
{
	Default
	{
		Scale 0.0;
		+FLOAT
		+NOGRAVITY
		+NOINTERACTION
		+NOBLOCKMAP
	}

	bool b1, closerThan512;
	Array<Actor> brightLights;

	override void beginPlay()
	{
		super.beginPlay();
		changeStatNum(STAT_LightSensor);
	}

	override void postBeginPlay()
	{
		int SPECIAL = 3;
		bool rl_debug = rlClass_CVars.get().rl_debug.GetBool();

		let maxDist = 1536 ** 2;
		let it = ThinkerIterator.create("BrightLight", STAT_BrightLight);

		Actor act;
		while (act = Actor(it.next()))
		{
			if (act == self || distance2DSquared(act) > maxDist)
			{
				continue;
			}

			brightLights.push(act);
		}

		let closestLight = getClosestLightDist(false);

		closerThan512 = closestLight <= 512;

		int ls = args[SPECIAL];
		if (ls != 10 && ls != 11 && ls != 12 && ls !=13) // door
		{
			b1 = closestLight <= 32;
		}
		else
		{
			b1 = false;
		}
		if (rl_debug)
		{
			self.Scale = (0.5, 0.5);
		}
	}

	override void tick()
	{
		// freed itself
		if (!checkNoDelay())
			return;

		// Advance the state
		if (tics != -1)
		{
			if (tics > 0)
			{
				tics--;
			}

			while (!tics)
			{
				// mobj was removed
				if (!setState (curState.nextState))
				{
					return;
				}
			}
		}
	}

	double getClosestLightDist(bool checkSight)
	{
		let maxDist = 1536 ** 2;

		double closest = double.infinity;
		for (int i = brightLights.Size () - 1; i >= 0; i--)
		{
			let light = brightLights[i];
			let thisDist = distance2DSquared(light);

			if (thisDist >= closest)
			{
				continue;
			}

			if (checkSight && !checkSight(light, SF_IGNOREVISIBILITY | SF_IGNOREWATERBOUNDARY))
			{
				continue;
			}

			closest = thisDist;
		}

		if (closest > maxDist)
		{
			return double.infinity;
		}

		return sqrt(closest);
	}

	int proximityLight()
	{
		let rl_Cvar = rlClass_Cvars.get();
		
		int SECTORID1 = 0;
		int SECTORID2 = 1;
		int DARKNESS = 2;
		int SPECIAL = 3;

		int minlight = args[DARKNESS];
		int light = rl_Cvar.rl_min.GetInt();

		if (closerThan512 || b1)
		{
			minlight = Level.Sectors[args[SECTORID1]].FindMinSurroundingLight(rl_Cvar.rl_max.GetInt() - rl_Cvar.rl_min.GetInt());
		}

		int valAdd;
		int maxAdd;

		if (b1) // doors < 32
		{
			valAdd = 64;
			maxAdd = 16;
		}
		else
		{
			let closestDist = getClosestLightDist(true);

			if (closestDist <= 64) // added
			{
				valAdd = 80;
				maxAdd = 32;
			}
			if (closestDist <= 128)
			{
				valAdd = 64;
				maxAdd = 16;
			}
			else if (closestDist <= 192) // added
			{
				valAdd = 48;
				maxAdd = 8;
			}
			else if (closestDist <= 256)
			{
				valAdd = 32;
				maxAdd = 4;
			}
			else if (closestDist <= 320) // added
			{
				valAdd = 16;
				maxAdd = 2;
			}
			else if (closestDist <= 384) // added
			{
				valAdd = 0;
				maxAdd = 0;
			}
			else if (closestDist <= 448) // added
			{
				valAdd = -16;
				maxAdd = -2;
			}
			else if (closestDist <= 512)
			{
				valAdd = -32;
				maxAdd = -4;
			}
			else if (closestDist <= 1024)
			{
				valAdd = -64;
				maxAdd = -16;
			}
			else if (closestDist <= 1536)
			{
				valAdd = -96;
				maxAdd = -32;
			}
			else
			{
				return light;
			}
		}

		return clamp(int((rl_Cvar.rl_max.GetInt() - rl_Cvar.rl_dynlux.GetInt() + valAdd + minlight) * 0.5 + 0.5), rl_Cvar.rl_min.GetInt(), rl_Cvar.rl_max.GetInt() - rl_Cvar.rl_dynlux.GetInt() + maxAdd);
	}
	States
	{
	Spawn:
		PVIS A 52 Bright
		{
			int SECTORID1 = 0;
			int SECTORID2 = 1;
			int DARKNESS = 2;
			int SPECIAL = 3;

			let rl_Cvar = rlClass_Cvars.get();
		
			double x = abs(self.pos.x - target.pos.x);
			double y = abs(self.pos.y - target.pos.y);
			
			if (x < rl_Cvar.rl_ldist.GetFloat() || y < rl_Cvar.rl_ldist.GetFloat())
			{
				int light = proximityLight();
				Level.Sectors[args[SECTORID1]].SetLightLevel(clamp(light + rl_Cvar.rl_add.GetInt(), 0, 255));

				if (args[SECTORID2] != -1)
				{
					Level.Sectors[args[SECTORID2]].SetLightLevel(clamp(light + rl_Cvar.rl_add.GetInt(), 0, 255));
				}
			}
			else if (x > rl_Cvar.rl_pdist.GetFloat() || y > rl_Cvar.rl_pdist.GetFloat())
			{
				SetStateLabel("Pause");
			}
			else if (x > rl_Cvar.rl_ddist.GetFloat() || y > rl_Cvar.rl_ddist.GetFloat())
			{
				SetStateLabel("Dormant");
			}
		}
		Loop;

	Pause:
		TNT1 A 520;
		TNT1 A 1
		{
			let rl_Cvar = rlClass_Cvars.get();

			double x = abs(self.pos.x - target.pos.x);
			double y = abs(self.pos.y - target.pos.y);
			if (x > rl_Cvar.rl_pdist.GetFloat() || y > rl_Cvar.rl_pdist.GetFloat())
			{
				SetStateLabel("Pause");
			}
			else if (x > rl_Cvar.rl_ddist.GetFloat() || y > rl_Cvar.rl_ddist.GetFloat())
			{
				SetStateLabel("Dormant");
			}
		}
		Goto Spawn;
	Dormant:
		TNT1 A 1040;
		TNT1 A 1
		{
			let rl_Cvar = rlClass_Cvars.get();

			double x = abs(self.pos.x - target.pos.x);
			double y = abs(self.pos.y - target.pos.y);
			if (x > rl_Cvar.rl_ddist.GetFloat() || y > rl_Cvar.rl_ddist.GetFloat())
			{
				SetStateLabel("Dormant");
			}
			else if (x > rl_Cvar.rl_pdist.GetFloat() || y > rl_Cvar.rl_pdist.GetFloat())
			{
				SetStateLabel("Pause");
			}
		}
		Goto Spawn;
	}
}

