// relightingevents.zs

#include "relighting_lib.zs"
#include "relighting_postprocessor.zs"
#include "relighting_actors.zs"

// relighting event handler
class relighting_EventHandler : EventHandler
{
	PlayerInfo player;
	int counter;
	int mintime;
	int maxtime;
	bool bDilate;
	int num_markers;
	int num_sensors;
	int num_lights;

/////////////////////////////////////////////////////////////////////////////////////////
	override void WorldTick()
	{
		let rl_Cvar = rlClass_Cvars.get();
		if (rl_Cvar.rl_default.GetBool())
		{
			rl_Cvar.setDefaults();
		}

		if (rl_Cvar.rl_enabled.GetBool() && rl_Cvar.rl_plr.GetInt() > 0)
		{
			if (counter == 0)
			{
				Sector sec = player.mo.CurSector;
				int seclight = sec.lightlevel;
				bool brightlights = false;
				
				if (seclight >= rl_Cvar.rl_max.GetInt() - rl_Cvar.rl_lux.GetInt() && bDilate && perceivedLight(sec.ColorMap.LightColor) > 128)
				{
					mintime = 0;
					maxtime++;
					if (maxtime == rl_Cvar.rl_plr.GetInt())
					{
						bDilate = false;
						if (rl_Cvar.rl_hudicon.GetBool())
						{
							player.mo.ACS_NamedExecuteAlways("constrictedPupil");
						}
						maxtime = 0;
						CVar.FindCvar("rl_dynlux").SetInt(rl_Cvar.rl_lux.GetInt());
					}
					else
					{
						if (rl_Cvar.rl_hudicon.GetBool())
						{
							player.mo.ACS_NamedExecuteAlways("normalPupil");
						}
					}
				}
				else if (seclight < rl_Cvar.rl_max.GetInt() - rl_Cvar.rl_lux.GetInt() - 16 && !bDilate)
				{
					if (rl_Cvar.rl_gldefs.GetBool())
					{
						brightlights = rlClass_Utils.checkProximitySimple(player.mo, "BrightLightActor", 1024, STAT_LightActor, true);
					}
					if (!brightlights)
					{
						maxtime = 0;
						mintime++;
						int rl_step = 4;
						if (mintime == 1)
						{
							// darkness is how dark beyond rl_lux plus 16
							// rl_step recovery is 1/4 of darkness
							
							int darkness = rl_Cvar.rl_lux.GetInt() - (rl_Cvar.rl_max.GetInt() - seclight) + 16;
							rl_step = int(darkness * 0.25 + 0.5);
							CVar.FindCvar("rl_dynlux").SetInt(clamp(int(seclight - darkness + 0.5), 0, 255));
						}
						else
						{
							if (rl_Cvar.rl_debug.GetBool())
							{
								console.printf("Adjusting %d %d", mintime, rl_Cvar.rl_dynlux.GetInt());
							}
							int new_lux = clamp(rl_Cvar.rl_dynlux.GetInt() - rl_step, rl_Cvar.rl_lux.GetInt(), 255);
							CVar.FindCvar("rl_dynlux").SetInt(new_lux);
							if (mintime == rl_Cvar.rl_plr.GetInt())
							{
								bDilate = true;
								if (rl_Cvar.rl_hudicon.GetBool())
								{
									player.mo.ACS_NamedExecuteAlways("dilatedPupil");
								}
								mintime = 0;
								CVar.FindCvar("rl_dynlux").SetInt(rl_Cvar.rl_lux.GetInt());
							}
						}
					}
				}
				counter = 38 * 1;
			}
			else
			{
				counter--;
			}
		}
	}
	void setPlayer(PlayerEvent e)
	{
		player = players[e.PlayerNumber];
		counter = 0;
		maxtime = 0;
		mintime = 0;
		bDilate = true;
		int rl_lux = CVar.FindCvar("rl_lux").GetInt();
		CVar.FindCvar("rl_dynlux").SetInt(rl_lux);

		int rl_plr = CVar.FindCvar("rl_plr").GetInt();
	}
	override void PlayerRespawned(PlayerEvent e)
	{
		setPlayer(e);
	}
	override void PlayerSpawned(PlayerEvent e)
	{
		setPlayer(e);
	}
/////////////////////////////////////////////////////////////////////////////////////////
/*
	override void WorldLineActivated(WorldEvent e)
	{
		Line lin = e.ActivatedLine;
		int special = lin.special;
		if (special == 250)
		{
			int tag = lin.Args[0];
			console.printf("%d", tag);

			let Tag_Finder = Level.CreateSectorTagIterator(tag);
			int i;
			while ((i = Tag_Finder.Next()) != -1)
			{
				// middle of the donut
				Sector sec = Level.Sectors[i];
				Sector fsec = sec.Lines[0].frontsector;
				Sector bsec = sec.Lines[0].backsector;

				if (fsec && sec != fsec)
				{
					fsec.SetFogDensity(0);
				}
				if (bsec && sec != bsec)
				{
					bsec.SetFogDensity(0);
				}
			}
		}
	}
*/
/////////////////////////////////////////////////////////////////////////////////////////
	override void WorldLoaded(WorldEvent e)
	{
		let rl_Cvar = rlClass_Cvars.get();

		int FLOOR = 0;
		int CEILING = 1;
		int INNER = 0;

		int SECTORID1 = 0;
		int SECTORID2 = 1;
		int DARKNESS = 2;
		int SPECIAL = 3;

		num_markers = 0;
		num_sensors = 0;
		num_lights = 0;

		Texman texture;
		if (!rl_Cvar.rl_enabled.GetBool())
		{
			return;
		}

		///////////////////////////////////////////////
		// read lump data

		let rl_flats = new("rlClass_WadLump");
		rl_flats.get("rl_flats");

		let rl_textures = new("rlClass_WadLump");
		rl_textures.get("rl_textures");

		let rl_gldefs = new("rlClass_WadLump");
		rl_gldefs.get("rl_gldefs");

		if (rl_Cvar.rl_reflections.GetBool())
		{
			let rl_reflections = new("rlClass_WadLump");
			rl_reflections.get("rl_reflections");
			string rl_tid = CVar.FindCvar("rl_tid").GetString();

			Array <string> tid_data;
			rl_tid.Split(tid_data, "|");
			for (int i = 0; i < tid_data.Size()-1; i++)
			{
				int tindex = rl_reflections.names.Find(tid_data[i]);
				if (tindex != rl_reflections.names.Size())
				{
					int tid = tid_data[i + 1].ToInt(10);
					int strength = rl_reflections.colors[tindex].ToInt(10);
					Sector_SetPlaneReflection (tid, strength, 0);
				}
			}
		}

		rlClass_Color sec_colors;
		rlClass_Color dl_colors;

		sec_colors.initialize();
		dl_colors.initialize();

		for(int i = 0; i < Level.Sectors.Size(); i++)
		{
			class<Actor> cname;
			Sector sec = Level.Sectors[i];
			SecPlane floorplane = sec.FloorPlane;
			SecPlane ceilingplane = sec.CeilingPlane;

			sec_colors.clear();
			bool bfloor = false;
			bool bceiling = false;
			bool bwalls = false;
			bool bsky = false;
			
			sec.SetLightLevel(rl_Cvar.rl_min.GetInt());
			int bright = 0;

			///////////////////////////////////////////////
			// find light and color sources

			// floors
			textureid floorid = Level.Sectors[i].GetTexture(FLOOR);
			int findex = rl_flats.names.Find(texture.GetName(floorid));
			if (findex != rl_flats.names.Size())
			{
				sec_colors.add(Color(rl_flats.colors[findex]));
				bfloor = true;
			}

			// ceilings
			textureid ceilingid = sec.GetTexture(CEILING);
			int cindex = rl_flats.names.Find(texture.GetName(ceilingid));
			if (cindex != rl_flats.names.Size())
			{
				sec_colors.add(Color(rl_flats.colors[cindex]));
				bceiling = true;
			}

			// sky flats are in textures...
			if (!bceiling && texture.GetName(ceilingid).IndexOf("_SKY") != -1)
			{
				textureid ceilingid = Level.SkyTexture1;
				int cindex = rl_flats.names.Find(texture.GetName(ceilingid));
				if (cindex != rl_flats.names.Size())
				{
					sec_colors.add(Color(rl_flats.colors[cindex]));
					bceiling = true;
					bsky = true;
				}
			}

			// walls
			for (int ii = 0; ii < sec.Lines.Size(); ii++)
			{
				Line lin = sec.Lines[ii];

				Side sidedef = lin.Sidedef[INNER];
			
				textureid topid = sidedef.GetTexture(sidedef.Top);
				textureid midid = sidedef.GetTexture(sidedef.Mid);
				textureid bottomid = sidedef.GetTexture(sidedef.Bottom);

				int tindex = rl_textures.names.Find(texture.GetName(topid));
				int mindex = rl_textures.names.Find(texture.GetName(midid));
				int bindex = rl_textures.names.Find(texture.GetName(bottomid));
				int tsize = rl_textures.names.Size();

				if (tindex != tsize)
				{
					sec_colors.add(Color(rl_textures.colors[tindex]));
					bwalls = true;
				}
				else if (mindex != tsize)
				{
					sec_colors.add(Color(rl_textures.colors[mindex]));
					bwalls = true;
				}
				else if (bindex != tsize)
				{
					sec_colors.add(Color(rl_textures.colors[bindex]));
					bwalls = true;
				}

			}
			if (sec_colors.colors.Size() > 0)
			{
				sec.SetColor(sec_colors.avg());
			}

			/////////////////////////////////////////////////////////////////////////////////////////////////////////
			// find target lines

			Array<int> twosidedLines;
			Array<int> doorLines;
			for (int ii = 0; ii < sec.Lines.Size(); ii++)
			{
				Line lin = sec.Lines[ii];
				int ls = lin.special;
				int f = lin.flags;

				if (ls == 10 || ls == 11 || ls == 12 || ls == 13) // door
				{
					doorLines.Push(ii);
				}
				else if (f & Line.ML_TWOSIDED)
				{
					twosidedLines.Push(ii);
				}
			}
			/////////////////////////////////////////////////////////////////////////////////////////////////////////
			// place BrightLight actors, adjust color and brightness

			if (bfloor || bceiling || bwalls)
			{
				int maxlights;
				if (rl_Cvar.rl_sources.GetInt() == 0)
				{
					maxLights = clamp(int(polygonArea(sec) / 4096), 1, 3); // performance
				}
				else
				{
					maxLights = clamp(int(polygonArea(sec) / 2048 + 0.5), 1, 4); // quality
				}
			
				for (int ii = 0; ii < doorLines.Size(); ii++)
				{
					if (maxLights == 0)
					{
						break;
					}
					Actor ptr = spawnLight("BrightLight", sec.Lines[doorLines[ii]], 8.0);
					if (ptr)
					{
						bright = clamp(bright++, 1, 4);
						num_markers++;
						maxLights--;
					}
				}
				int step = twosidedLines.Size() > 25 ? 2 : 1;
				for (int ii = 0; ii < twosidedLines.Size(); ii+=step)
				{
					if (maxLights == 0)
					{
						break;
					}
					Actor ptr = spawnLight("BrightLight", sec.Lines[twosidedLines[ii]], 128.0);
					if (ptr)
					{
						bright = clamp(bright++, 1, 4);
						num_markers++;
						maxLights--;
					}
				}

				/////////////////////////////////////////////////////////////////////////////////////////////////////////
				// final adjustment to sector brightness

				if (bright > 0)
				{
					bright = clamp(sec.Lines.Size() / bright, 1, 4);
					bright = clamp(bright + int(polygonArea(sec) / 1024 + 0.5), bright, 5);
				}
				if (bfloor)
				{
					bright = clamp(bright++, bright, 3);
				}
				if (bceiling)
				{
					bright = clamp(bright++, bright, 5);
				}
				if (perceivedLight(sec.ColorMap.LightColor) > 128)
				{
					bright = clamp(bright++, bright, 5);
				}
				if (perceivedLight(sec.ColorMap.LightColor) <= 128)
				{
					bright = clamp(bright--, 1, bright);
				}
				int lightbefore = sec.lightlevel;
				int fog = 0;
				
				switch (bright)
				{
					case 1:
						sec.SetLightLevel(clamp(rl_Cvar.rl_max.GetInt() - rl_Cvar.rl_min.GetInt() * 4, rl_Cvar.rl_min.GetInt(), rl_Cvar.rl_max.GetInt()));
						fog = clamp(rl_Cvar.rl_fog.GetInt() * 4, 0, 255);
						break;
					case 2:
						sec.SetLightLevel(clamp(rl_Cvar.rl_max.GetInt() - rl_Cvar.rl_min.GetInt() * 3, rl_Cvar.rl_min.GetInt(), rl_Cvar.rl_max.GetInt()));
						fog = clamp(rl_Cvar.rl_fog.GetInt() * 2, 0, 255);
						break;
					case 3:
						sec.SetLightLevel(clamp(rl_Cvar.rl_max.GetInt() - rl_Cvar.rl_min.GetInt() * 2, rl_Cvar.rl_min.GetInt(), rl_Cvar.rl_max.GetInt()));
						fog = clamp(rl_Cvar.rl_fog.GetInt() * 1, 0, 255);
						break;
					case 4:
						sec.SetLightLevel(clamp(rl_Cvar.rl_max.GetInt() - rl_Cvar.rl_min.GetInt() * 1, rl_Cvar.rl_min.GetInt(), rl_Cvar.rl_max.GetInt()));
						fog = clamp(int(rl_Cvar.rl_fog.GetInt() * 0.5 + 0.5), 0, 255);
						break;
					case 5:
					case 6:
					case 7:
					case 8:
						sec.SetLightLevel(rl_Cvar.rl_max.GetInt());
						fog = clamp(int(rl_Cvar.rl_fog.GetInt() * 0.25 + 0.5), 0, 255);
						break;
					default:
						fog = 0;
						break;
				}
				if (bright && rl_Cvar.rl_fog.GetInt() > 0 && polygonArea(sec) < rl_Cvar.rl_fogarea.GetInt())
				{
					color secColor = sec.ColorMap.LightColor;
					// sec.SetFade(secColor);
					// sec.SetFogDensity(fog);
				}
				if (rl_Cvar.rl_debug.GetBool() && lightbefore != sec.lightlevel)
				{
					console.printf("%d before: %d after: %d", i, lightbefore, sec.lightlevel);
					//sec.SetColor("#9932cc");
				}
			}
			/////////////////////////////////////////////////////////////////////////////////////////////////////////
			// place LightSensor actor

			else
			{
				int maxLights = 1;

				// doors
				for (int ii = 0; ii < doorLines.Size(); ii++)
				{
					if (maxLights == 0)
					{
						break;
					}
					Actor ptr = spawnLight("LightSensor", sec.Lines[doorLines[ii]], 8.0);
					if (ptr)
					{
						num_sensors++;
						maxLights--;
						ptr.target = player.mo;
						ptr.args[SECTORID1] = i;
						ptr.args[SPECIAL] = sec.Lines[doorLines[ii]].special;
						ptr.args[DARKNESS] = sec.FindMinSurroundingLight(rl_Cvar.rl_max.GetInt() - rl_Cvar.rl_min.GetInt());
					}
				}
				// back sector is dark
				for (int ii = 0; ii < twosidedLines.Size(); ii++)
				{
					if (maxLights == 0)
					{
						break;
					}
					Sector bsec = sec.Lines[twosidedLines[ii]].backsector;
					if (bsec)
					{
						if (bsec.lightlevel== rl_Cvar.rl_min.GetInt())
						{
							Actor ptr = spawnLight("LightSensor", sec.Lines[twosidedLines[ii]], 8.0);
							if (ptr)
							{
								num_sensors++;
								maxLights--;
								ptr.target = player.mo;
								ptr.args[SECTORID1] = i;
								ptr.args[SPECIAL] = sec.Lines[twosidedLines[ii]].special;
								ptr.args[DARKNESS] = sec.FindMinSurroundingLight(rl_Cvar.rl_max.GetInt() - rl_Cvar.rl_min.GetInt());
							}
						}
					}
				}
				// back sector is bright
				for (int ii = 0; ii < twosidedLines.Size(); ii++)
				{
					if (maxLights == 0)
					{
						break;
					}
					Sector bsec = sec.Lines[twosidedLines[ii]].backsector;
					if (bsec)
					{
						if (bsec.lightlevel > rl_Cvar.rl_min.GetInt())
						{
							Actor ptr = spawnLight("LightSensor", sec.Lines[twosidedLines[ii]], 8.0);
							if (ptr)
							{
								num_sensors++;
								maxLights--;
								ptr.target = player.mo;
								ptr.args[SECTORID1] = i;
								ptr.args[SPECIAL] = sec.Lines[twosidedLines[ii]].special;
								ptr.args[DARKNESS] = sec.FindMinSurroundingLight(rl_Cvar.rl_max.GetInt() - rl_Cvar.rl_min.GetInt());
							}
						}
					}
				}
			}
			///////////////////////////////////////////////
			// dynamic light sources (beta)
			for (int ii = 0; ii < sec.Lines.Size(); ii++)
			{
				Line lin = sec.Lines[ii];

				Side sidedef = lin.Sidedef[INNER];
			
				textureid topid = sidedef.GetTexture(sidedef.Top);
				textureid midid = sidedef.GetTexture(sidedef.Mid);
				textureid bottomid = sidedef.GetTexture(sidedef.Bottom);

				int tindex = rl_textures.names.Find(texture.GetName(topid));
				int mindex = rl_textures.names.Find(texture.GetName(midid));
				int bindex = rl_textures.names.Find(texture.GetName(bottomid));
				int tsize = rl_textures.names.Size();

				///////////////////////////////////////////////
				// dynamic light sources (beta)

				if (rl_Cvar.rl_gldefs.GetBool())
				{
					vector2 v2 = getMiddle(lin);
					double x = v2.x;
					double y = v2.y;
					double xylength = lin.delta.length();

					double z = floorplane.ZAtPoint((x,y));
					double zheight = ceilingplane.ZAtPoint((x,y)) - floorplane.ZAtPoint((x,y));

					// windows
					int flight = sec.lightlevel;
					int blight = flight;
					Sector bsec = lin.backsector;
					if (bsec)
					{
						blight = bsec.lightlevel;
					}
					int f2b = flight - blight;
					if ((bsky || (flight > 224 && f2b > rl_Cvar.rl_min.GetInt() && 
					f2b < rl_Cvar.rl_lux.GetInt())) && rl_Cvar.rl_glwindows.GetBool())
					{
						Sector bsec = lin.backsector;
						if (bsec)
						{
							double z = ZatMiddle(lin);

							textureid bsec_ceilingid = bsec.GetTexture(CEILING);
							if (texture.GetName(bsec_ceilingid).IndexOf("_SKY1") == -1)
							{
								int f = lin.flags;
								int peg_top = f & (Line.ML_TWOSIDED + Line.ML_DONTPEGTOP);
								int peg_bottom = f & (Line.ML_TWOSIDED + Line.ML_DONTPEGBOTTOM);
								int peg_both = f & (Line.ML_TWOSIDED + Line.ML_DONTPEGTOP + Line.ML_DONTPEGBOTTOM);
								Side sidedef = lin.Sidedef[INNER];
							
								textureid midid = sidedef.GetTexture(sidedef.Mid);
								if (midid.isNull() && (f | Line.ML_DONTPEGTOP || f | Line.ML_DONTPEGBOTTOM))
								{
									spawnDynamicLight(lin, sec.ColorMap.LightColor, "MIDDLE");
								}
							}
						}
					}

					// textures
					int dl_size = rl_gldefs.names.Size();
					if (!topid.IsNull())
					{
						int dl_tindex = rl_gldefs.names.Find(texture.GetName(topid));
						if (dl_tindex != dl_size)
						{
							spawnDynamicLight(lin, Color(rl_gldefs.colors[dl_tindex]), "TOP");
						}
					}
					if (!midid.IsNull())
					{
						int dl_mindex = rl_gldefs.names.Find(texture.GetName(midid));
						if (dl_mindex != dl_size)
						{
							spawnDynamicLight(lin, Color(rl_gldefs.colors[dl_mindex]), "MIDDLE");
						}
					}
					if (!bottomid.IsNull())
					{
						int dl_bindex = rl_gldefs.names.Find(texture.GetName(bottomid));
						if (dl_bindex != dl_size)
						{
							spawnDynamicLight(lin, Color(rl_gldefs.colors[dl_bindex]), "BOTTOM");
						}
					}
				}
			}
		}
		
		if (rl_Cvar.rl_debug.GetBool())
		{
			console.printf("light markers: %d, sensors: %d, dynamic lights: %d", num_markers, num_sensors, num_lights);
		}
	}

/////////////////////////////////////////////////////////////////////////////////////////
// functions

	void spawnDynamicLight(Line lin, Color dl_color, string side = "MIDDLE")
	{
		let rl_Cvar = rlClass_Cvars.get();

		vector2 v2 = getMiddle(lin);
		double x = v2.x;
		double y = v2.y;
		double xylength = lin.delta.length();
	
		Sector sec = lin.frontsector;
		Actor ptr;
		double gl_light = perceivedLight(dl_color);

		// if (gl_light > rl_Cvar.rl_max.GetInt() - rl_Cvar.rl_lux.GetInt())
		if (gl_light > 223)
		{
			ptr = spawnLight("BrightLightActor", lin, 32.0, side);
		}
		else
		{
			ptr = spawnLight("LightActor", lin, 16.0, side);
		}
	
		if (ptr)
		{
			num_lights++;

			int sec_light = sec.lightlevel;
			int gl_type;

			double area = getArea(lin, side) / 4096;

			if (gl_light > rl_Cvar.rl_max.GetInt() - rl_Cvar.rl_lux.GetInt())
//			if (gl_light >= 128)
			{
				gl_type = DynamicLight.SectorLight;

			}
			else
			{
				gl_type = DynamicLight.PointLight;
			}

			int flags = DYNAMICLIGHT.LF_ATTENUATE;

			if (gl_light < 128 && sec_light < 128)
			{
				flags = flags | DYNAMICLIGHT.LF_DONTLIGHTACTORS;
			}
			if (gl_light < 64)
			{
				flags = flags | DYNAMICLIGHT.LF_SUBTRACTIVE;
			}
		
			
			ptr.A_AttachLight("Light", gl_type, dl_color,
			int(area * rl_Cvar.rl_glstrength.GetFloat() + 0.5), 0, flags);
		}
	}

	Actor spawnLight(class<Actor> cname, Line lin, double dist, string side = "MIDDLE")
	{
		let rl_Cvar = rlClass_Cvars.get();

		int SECTORID1 = 0;
		int SECTORID2 = 1;

		vector2 v2 = getMiddle(lin);
		double x = v2.x;
		double y = v2.y;
		Sector sec = lin.frontsector;

		double z;

		// hacky... should fix this

		int ls = lin.special;
		if (ls == 10 || ls == 11 || ls == 12 || ls == 13) // door
		{
			z = ZatMiddle(lin, side) + 32;
		}
		else
		{
			z = ZatMiddle(lin, side);
		}
		if (z == double.infinity)
		{
			return null;
		}

		int statNum;
		if (cname is 'LightSensor')
			statNum = STAT_LightSensor;
		else if (cname is 'BrightLight')
			statNum = STAT_BrightLight;
		else if (cname is 'BrightLightActor')
			statNum = STAT_LightActor;
		else if (cname is 'LightActor')
			statNum = STAT_LightActor;
		else
			throwAbortException("spawnLight: Invalid cname \"%s\".", cname.GetClassName ());

		Actor ptr = Actor.Spawn(cname, (x,y,z));
		if (ptr)
		{
			if (cname is 'BrightLightActor' || cname is 'LightActor')
			{
				if (rlClass_Utils.checkProximitySimple(ptr, cname, dist, statNum, true, true))
				{
					ptr.destroy();
				}
			}
			else if (rlClass_Utils.checkProximitySimple(ptr, cname, dist, statNum, true))
			{
				if (cname == "LightSensor")
				{
					Actor dupeptr = null;
					let LightSensor_Finder = ThinkerIterator.Create("LightSensor", STAT_LightSensor);
					LightSensor lspoint;

					while (lspoint = LightSensor(LightSensor_Finder.Next()))
					{
						if (lspoint != ptr && lspoint.pos.x == ptr.pos.x && lspoint.pos.y == ptr.pos.y)
						{
							if (rl_Cvar.rl_debug.GetBool())
							{
								console.printf("Dupe %d %d", lspoint.pos.x, lspoint.pos.y);
							}
							dupeptr = lspoint;
							break;
						}
					}
					if (dupeptr)
					{
						ptr.args[SECTORID2] = dupeptr.args[SECTORID1];
						dupeptr.destroy();
						num_sensors--;
					}
				}
				else if (cname is 'BrightLight')
				{
					ptr.destroy();
				}
			}
		}
		return ptr;
	}
	vector2 getMiddle(Line lin)
	{
		double x = (lin.V1.P.x + lin.V2.P.x) * .5;
		double y = (lin.V1.P.y + lin.V2.P.y) * .5;
		return (x, y);
	}

	double getArea(Line lin, string side = "MIDDLE")
	{
		int INNER = 0;
		vector2 v2 = getMiddle(lin);
		Sector sec = lin.frontsector;
		SecPlane floorplane = sec.FloorPlane;
		SecPlane ceilingplane = sec.CeilingPlane;

		double xylength = lin.delta.length();

		int f = lin.flags;
		int peg_top = f & (Line.ML_TWOSIDED + Line.ML_DONTPEGTOP);
		int peg_bottom = f & (Line.ML_TWOSIDED + Line.ML_DONTPEGBOTTOM);
		int peg_both = f & (Line.ML_TWOSIDED + Line.ML_DONTPEGTOP + Line.ML_DONTPEGBOTTOM);
		Side sidedef = lin.Sidedef[INNER];
	
		textureid midid = sidedef.GetTexture(sidedef.Mid);

		if (midid.isNull() && (f | Line.ML_DONTPEGTOP || f | Line.ML_DONTPEGBOTTOM))
		{
			Sector bsec = lin.backsector;
			if (bsec)
			{
				SecPlane bsec_floorplane = bsec.FloorPlane;
				SecPlane bsec_ceilingplane = bsec.CeilingPlane;
			
				if (side == "TOP")
				{
					return (xylength * (ceilingplane.ZAtPoint((v2.x, v2.y)) + bsec_ceilingplane.ZAtPoint((v2.x, v2.y))) * 0.5);
				}
				if (side == "MIDDLE")
				{
					return (xylength * (bsec_ceilingplane.ZAtPoint((v2.x, v2.y)) + bsec_floorplane.ZAtPoint((v2.x, v2.y))) * 0.5);
				}
				if (side == "BOTTOM")
				{
					return (xylength * (floorplane.ZAtPoint((v2.x, v2.y)) + bsec_floorplane.ZAtPoint((v2.x, v2.y))) * 0.5);
				}
			}
		}
		return (xylength * (ceilingplane.ZAtPoint((v2.x, v2.y)) + floorplane.ZAtPoint((v2.x, v2.y))) * 0.5);
	}

	double ZatMiddle(Line lin, string side = "MIDDLE")
	{
		int INNER = 0;
		Sector sec = lin.frontsector;
		SecPlane floorplane = sec.FloorPlane;
		SecPlane ceilingplane = sec.CeilingPlane;

		int ls = lin.special;
		bool bdoor = (ls == 10 || ls == 11 || ls == 12 || ls == 13); // door hack

		vector2 v2 = getMiddle(lin);
		int f = lin.flags;
		int peg_top = f & (Line.ML_TWOSIDED + Line.ML_DONTPEGTOP);
		int peg_bottom = f & (Line.ML_TWOSIDED + Line.ML_DONTPEGBOTTOM);
		int peg_both = f & (Line.ML_TWOSIDED + Line.ML_DONTPEGTOP + Line.ML_DONTPEGBOTTOM);
		Side sidedef = lin.Sidedef[INNER];
	
		textureid midid = sidedef.GetTexture(sidedef.Mid);
		if (midid.isNull() && (f | Line.ML_DONTPEGTOP || f | Line.ML_DONTPEGBOTTOM))
		{
			Sector bsec = lin.backsector;
			if (bsec)
			{
				SecPlane bsec_floorplane = bsec.FloorPlane;
				SecPlane bsec_ceilingplane = bsec.CeilingPlane;
			
				if (side == "TOP")
				{
					if (ceilingplane.ZAtPoint((v2.x, v2.y)) == bsec_ceilingplane.ZAtPoint((v2.x, v2.y)) && !bdoor)
					{
						return double.infinity;
					}
					else
					{
						return (ceilingplane.ZAtPoint((v2.x, v2.y)) + bsec_ceilingplane.ZAtPoint((v2.x, v2.y))) * 0.5;
					}
				}
				if (side == "MIDDLE")
				{
					return (bsec_ceilingplane.ZAtPoint((v2.x, v2.y)) + bsec_floorplane.ZAtPoint((v2.x, v2.y))) * 0.5;
				}
				if (side == "BOTTOM")
				{
					if (floorplane.ZAtPoint((v2.x, v2.y)) == bsec_floorplane.ZAtPoint((v2.x, v2.y)))
					{
						return double.infinity;
					}
					else
					{
						return (floorplane.ZAtPoint((v2.x, v2.y)) + bsec_floorplane.ZAtPoint((v2.x, v2.y))) * 0.5;
					}
				}
			}
		}

		return (ceilingplane.ZAtPoint((v2.x, v2.y)) + floorplane.ZAtPoint((v2.x, v2.y))) * 0.5;
	}

	float polygonArea(Sector sec)
	{
		Array<int> X;
		Array<int> Y;
		for (int i = 0; i < sec.Lines.Size(); i++)
		{
			X.Push(int(sec.Lines[i].V1.P.x + 0.5));
			Y.Push(int(sec.Lines[i].V1.P.y + 0.5));
		}

		float area = 0.0;
		for (int i = 0; i < X.Size() - 1; i++)
		{
			area += (X[i] * Y[i+1]) - (X[i+1] * Y[i]);
		}
		area = abs(area) / 2.0 / 64.0;

		return area;
	}
	// https://stackoverflow.com/questions/596216/formula-to-determine-perceived-brightness-of-rgb-color
	double perceivedLight(color c)
	{
		let rl_Cvar = rlClass_Cvars.get();
		int rl_glluma = rl_Cvar.rl_glluma.GetInt();

		if (rl_glluma == 0)
		{
			// ITU BT.601 - Digital more weight to R and B
			return sqrt(0.299 * (c.r * c.r) + 0.587 * (c.g * c.g) + 0.114 * (c.b * c.b));
		}
		else
		{
			// ITU BT.709 - HDTV standard
			return sqrt(0.2126 * (c.r * c.r) + 0.7152 * (c.g * c.g) + 0.0722 * (c.b * c.b));
		}
	}
}

class relighting_OptionMenu : OptionMenu
{
	override void Init(Menu parent, OptionMenuDescriptor desc)
	{
		super.Init(parent, desc);
		DontDim = true;
		DontBlur = true;
	}
}
