const STAT_BrightLight = Thinker.STAT_USER + 1;
const STAT_LightSensor = Thinker.STAT_USER + 2;
const STAT_LightActor = Thinker.STAT_USER + 3;

// base classes
class rlClass_CVars : Thinker
{
	transient bool cvarsLoaded;
	
    transient CVar rl_default;
    transient CVar rl_enabled;
    transient CVar rl_gldefs;
    transient CVar rl_glwindows;
    transient CVar rl_reflections;
    transient CVar rl_hudicon;
    transient CVar rl_debug;

	transient CVar rl_max;
	transient CVar rl_min;
	transient CVar rl_lux;
	transient CVar rl_dynlux;
	transient CVar rl_add;
	transient CVar rl_alpha;
	transient CVar rl_sources;
	transient CVar rl_fog;
	transient CVar rl_fogarea;
	transient CVar rl_plr;
	transient CVar rl_glluma;

	transient CVar rl_glstrength;
	transient CVar rl_ldist;
	transient CVar rl_pdist;
	transient CVar rl_ddist;

	void initialize()
	{
		ChangeStatNum(Thinker.STAT_Info);
		initializeCvars();
	}
	
	void initializeCvars()
	{
		if (cvarsLoaded)
		{
			return;
		}

		rl_default = CVar.findCvar("rl_default");
		rl_enabled = CVar.findCvar("rl_enabled");
		rl_gldefs = CVar.findCvar("rl_gldefs");
		rl_glwindows = CVar.findCvar("rl_glwindows");
		rl_reflections = CVar.findCvar("rl_reflections");
		rl_hudicon = CVar.findCvar("rl_hudicon");
		rl_debug = CVar.findCvar("rl_debug");

		rl_max = CVar.findCvar("rl_max");
		rl_min = CVar.findCvar("rl_min");
		rl_lux = CVar.findCvar("rl_lux");
		rl_dynlux = CVar.findCvar("rl_dynlux");
		rl_add = CVar.findCvar("rl_add");
		rl_alpha = CVar.findCvar("rl_alpha");
		rl_sources = CVar.findCvar("rl_sources");
		rl_fog = CVar.findCvar("rl_fog");
		rl_fogarea = CVar.findCvar("rl_fogarea");
		rl_plr = CVar.findCvar("rl_plr");
		rl_glluma = CVar.findCvar("rl_glluma");

		rl_glstrength = CVar.findCvar("rl_glstrength");
		rl_ldist = CVar.findCvar("rl_ldist");
		rl_pdist = CVar.findCvar("rl_pdist");
		rl_ddist = CVar.findCvar("rl_ddist");
		
		cvarsLoaded = true;
	}

	static rlClass_CVars get()
	{
		let it = ThinkerIterator.create("rlClass_CVars", Thinker.STAT_Info);
		let ret = rlClass_CVars(it.next());

		if (!ret)
		{
			ret = new("rlClass_CVars");
			ret.initialize();
		}
		else
		{
			ret.initialize();
		}

		return ret;
	}

	static void setDefaults()
	{
		CVar.findCvar("rl_default").resetToDefault();
		CVar.findCvar("rl_enabled").resetToDefault();
		CVar.findCvar("rl_gldefs").resetToDefault();
		CVar.findCvar("rl_glwindows").resetToDefault();
		CVar.findCvar("rl_reflections").resetToDefault();
		CVar.findCvar("rl_hudicon").resetToDefault();
		CVar.findCvar("rl_debug").resetToDefault();

		CVar.findCvar("rl_max").resetToDefault();
		CVar.findCvar("rl_min").resetToDefault();
		CVar.findCvar("rl_lux").resetToDefault();
		CVar.findCvar("rl_dynlux").resetToDefault();
		CVar.findCvar("rl_add").resetToDefault();
		CVar.findCvar("rl_alpha").resetToDefault();
		CVar.findCvar("rl_sources").resetToDefault();
		CVar.findCvar("rl_fog").resetToDefault();
		CVar.findCvar("rl_fogarea").resetToDefault();
		CVar.findCvar("rl_plr").resetToDefault();
		CVar.findCvar("rl_glluma").resetToDefault();

		CVar.findCvar("rl_glstrength").resetToDefault();
		CVar.findCvar("rl_ldist").resetToDefault();
		CVar.findCvar("rl_pdist").resetToDefault();
		CVar.findCvar("rl_ddist").resetToDefault();
	}
}

class rlClass_WadLump
{
	Array<string> names;
	Array<string> colors;

	string RTrim(string s)
	{
		while (s.mid(s.length()-1, 1) == " ")
		{
			s = s.left(s.length() - 2);
		}
		return s;
	}
	void get(string s)
	{
		int fhnd = Wads.findLump(s, 0, Wads.GLOBALNAMESPACE);
		string flump = Wads.readLump(fhnd);
		Array <string> data;
		flump.split(data, "\r\n");

		for (int i = 0; i < data.Size(); i++)
		{
			if (data[i].Left(1) != ";")
			{
				Array<string> item;
				data[i].Split(item, ",");
				names.Push(RTrim(item[0]));
				colors.Push(RTrim(item[1]));
			}
		}
	}
}

struct rlClass_Color
{
	Array<color> colors;
	int rl_alpha;

	void initialize()
	{
		colors.clear();
		rl_alpha = CVar.findCVar("rl_alpha").getInt();
	}

	void clear()
	{
		colors.clear();
	}

	void add(color c)
	{
		colors.push(c);
	}

	color avg()
	{
		int r = 0;
		int g = 0;
		int b = 0;
		int colorsCount = colors.size();
		if (colorsCount > 0)
		{
			for (int j = 0; j < colorsCount; j++)
			{
				r += colors[j].r;
				g += colors[j].g;
				b += colors[j].b;
			}
			r /= colorsCount;
			g /= colorsCount;
			b /= colorsCount;

			if (r != 255 || g != 255 || b != 255)
			{
				r = clamp(r + rl_alpha, 0, 255);
				g = clamp(g + rl_alpha, 0, 255);
				b = clamp(b + rl_alpha, 0, 255);
			}
		}
		return Color(255, r, g, b);
	}
}

class rlClass_Utils : Actor
{
	static bool checkProximitySimple(Actor origin, class<Actor> actorName, float dist, int statNum, bool checkSight, bool checkZ = false)
	{
		dist *= dist;

		let it = ThinkerIterator.create(actorName, statNum);

		Actor act;
		while (act = Actor(it.next()))
		{
			if (act == origin || origin.distance2DSquared(act) > dist)
			{
				continue;
			}

			if (checkZ && origin.pos.z == act.pos.z)
			{
				return true;
			}

			if (checkSight && !origin.checkSight(act, SF_IGNOREVISIBILITY | SF_IGNOREWATERBOUNDARY))
			{
				continue;
			}

			return true;
		}

		return false;
	}
}